/*
 * OpenCL mask defines
 *
 * This file needs only one #define that is also needed by
 * non-OpenCL mask routines. The 'define' was copied here
 * to be used by Kernels ('*.cl') source code.
 *
 */

#ifndef _JOHN_OPENCL_MASK_H
#define _JOHN_OPENCL_MASK_H

// See also mask.h.
#define MASK_FMT_INT_PLHDR 		4

#endif
